use crate::{router::ReqResult, service::accounts::AuthSession};
use axum::response::Html;
use maud::{html, Markup, PreEscaped, DOCTYPE};

pub fn stylesheet(url: &str) -> Markup {
    html! {
        link rel="stylesheet" type="text/css" href=(url);
    }
}

pub fn head(page_title: &str) -> Markup {
    html! {
        (DOCTYPE)
        meta charset="utf-8" {}
        meta name="viewport" content="width=device-width, initial-scale=1" {}
        title { (page_title) " - lyrs" }
        (stylesheet("/static/style.css"));

        script src="https://unpkg.com/htmx.org@1.9.12" integrity="sha384-ujb1lZYygJmzgSwoxRggbCHcjc0rB2XoQrxeTUQyRjrOnlCoYta87iKBWq3EsdM2" crossorigin="anonymous" defer {}
    }
}

pub fn foot() -> Markup {
    html! {
        footer class="p-2 bg-mantle border-t-2 border-surface0 flex overflow-x-scroll mt-auto" {
            section {
                (PreEscaped("&copy; 2024 "))
                a class="underline text-mauve" href="https://lyte.dev" { "lytedev" }
            }
            section .ml-auto {("Made with ❤️")}
            " "
            a .underline.text-mauve href="/about" { "About" }
        }
    }
}

pub fn page(
    title: &str,
    content: Markup,
    auth_session: Option<AuthSession>,
) -> ReqResult<Html<String>> {
    let current_user = auth_session.map(|s| s.user).flatten();
    Ok(Html(
    html! {
        (head(title))
        body hx-boost="true" class="bg-bg text-text min-h-lvh flex flex-col font-sans overflow-x-hidden" {
            header class="drop-shadow border-b-2 border-surface0 bg-mantle flex overflow-x-scroll" {
                a class="flex p-2 text-3xl font-mono text-mauve opacity-80 hover:bg-mauve hover:text-bg" href="/" { "lyrs" }
                nav class="flex flex-1 justify-start" {
                    @if let Some(user) = current_user {
                        a class="flex items-center p-2 hover:bg-mauve hover:text-bg" href="/dashboard" {( user.username )}
                        a class="flex items-center p-2 hover:bg-mauve hover:text-bg" href="/accounts/logout" { "Logout" }
                    } @else {
                        a class="flex items-center p-2 hover:bg-mauve hover:text-bg" href="/accounts/login" { "Login" }
                        a class="flex items-center p-2 hover:bg-mauve hover:text-bg" href="/accounts/register" { "Register" }
                    }
                }
            }
            (content)
        }
        (foot())
    }.into_string()
    ))
}

pub fn center_hero_form(title: &str, content: Markup, subform: Markup) -> Markup {
    html! {
        section class="hero grow" {
            form class="flex flex-col gap-2 w-full max-w-sm" method="post" {
                header {
                    h1 class="pb-2 text-center text-xl" { (title) }
                }
                (content)
                input class="bg-mauve" value="Submit" type="submit";
            }
            (subform)
        }
    }
}

pub fn labelled_input(label: &str, input: Markup) -> Markup {
    html! {
        label class="flex flex-col" {
            (label)
            (input)
        }
    }
}
