use crate::router::ReqResult;
use crate::state::{Creds, State as AppState};
use crate::{partials::*, user::User};
use crate::{prelude::*, user};
use axum::extract::State;
use axum::http::StatusCode;
use axum::response::{Html, IntoResponse, Redirect};
use axum::Form;
use maud::html;
use std::convert::Infallible;
use tracing::instrument;

use axum::{
    routing::{get, post},
    Router,
};

pub fn router(state: AppState) -> Result<Router, Infallible> {
    Ok(Router::new()
        .route("/login", get(login))
        .route("/login", post(authenticate))
        .route("/register", get(register))
        .route("/register", post(create_user))
        .route("/logout", get(logout))
        .with_state(state))
}

async fn login(auth_session: Option<AuthSession>) -> impl IntoResponse {
    if auth_session.map(|s| s.user).flatten().is_some() {
        return Redirect::to("/dashboard").into_response();
    }
    let form = html! {
        (labelled_input("Username", html!{
            input class="input" type="text" name="username" autocomplete="username" required;
        }))
        (labelled_input("Password", html!{
            input class="input" type="password" name="password" autocomplete="current-password" required;
        }))
    };
    let subaction = html! {
        small class="mt-4" {
            "Need an account? "
            a href="/accounts/register" {"Get one"}
            "."
        }
    };
    page("login", center_hero_form("Login", form, subaction), None).into_response()
}

async fn register(auth_session: Option<AuthSession>) -> impl IntoResponse {
    if auth_session.map(|s| s.user).flatten().is_some() {
        return Redirect::to("/dashboard").into_response();
    }
    let form = html! {
        (labelled_input("Username", html!{
            input class="input" type="text" name="username" required;
        }))
        (labelled_input("Password", html!{
            input class="input" type="password" name="password" autocomplete="new-password" required;
        }))
    };
    let subaction = html! {
        small class="mt-4" {
            "Already have an account? "
            a href="/accounts/login" {"Login"}
            "."
        }
    };
    page("login", center_hero_form("Register", form, subaction), None).into_response()
}

pub type AuthSession = axum_login::AuthSession<AppState>;

#[instrument(skip(auth_session))]
async fn authenticate(
    mut auth_session: AuthSession,
    Form(creds): Form<Creds>,
) -> impl IntoResponse {
    let user = match auth_session.authenticate(creds.clone()).await {
        Ok(Some(user)) => user,
        Ok(None) => return StatusCode::UNAUTHORIZED.into_response(),
        Err(_) => return StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    };

    if auth_session.login(&user).await.is_err() {
        return StatusCode::INTERNAL_SERVER_ERROR.into_response();
    }

    Redirect::to("/dashboard").into_response()
}

#[instrument(skip(state))]
async fn create_user(
    State(state): State<AppState>,
    Form(creds): Form<Creds>,
) -> ReqResult<Html<String>> {
    let user = User::try_new(&creds.username, creds.password.expose_secret())?;
    let existing_user: Option<User> = state.db.get(User::tree(), &creds.username)?;

    // TODO: fail2ban?
    if existing_user.is_some() {
        // timing/enumeration attacks or something
        return Err(user::Error::UsernameExists(Box::new(creds.username)).into());
    }

    state
        .db
        .insert(User::tree(), &user.username, bincode::serialize(&user)?)?;

    Ok(Html(
        html! {
            ({&user.username}) " has been registered"
        }
        .into_string(),
    ))
}

pub async fn logout(mut auth_session: AuthSession) -> impl IntoResponse {
    match auth_session.logout().await {
        Ok(Some(_u)) => Redirect::to("/accounts/login?done").into_response(),
        Ok(None) => Redirect::to("/accounts/login?unauthenticated").into_response(),
        Err(_) => StatusCode::INTERNAL_SERVER_ERROR.into_response(),
    }
}
